# Copyright © 2015 Polyconseil SAS
# SPDX-License-Identifier: BSD-3-Clause
#

def swap_bytes(data):
    # Inverse byte order for DBC 0xAABBCCDD to 0xDDCCBBAA
    data_swapped = ''
    for i in range(len(data), 0, -8):
        data_swapped += data[i - 8:i]
    return data_swapped


def hex_ascii_to_bytes(hex_ascii_str):
    hex_ascii_list = []
    # Length in bytes
    length = int(len(hex_ascii_str) / 2)
    for i in range(length):
        hex_ascii_list.append(int(hex_ascii_str[i * 2:i * 2 + 2], 16))

    return length, bytes(hex_ascii_list)



# Copyright © 2015 Polyconseil SAS
# SPDX-License-Identifier: BSD-3-Clause
#

from .decode import (MESSAGE_MAX_LENGTH,
    message_decode, message_get_current_multiplexing_mode,
    signal_decode)
from .utils import swap_bytes, hex_ascii_to_bytes

__all__ = [
    'MESSAGE_MAX_LENGTH',
    'message_decode', 'message_get_current_multiplexing_mode',
    'signal_decode',
    'swap_bytes', 'hex_ascii_to_bytes'
]



# Copyright © 2015 Polyconseil SAS
# SPDX-License-Identifier: BSD-3-Clause
#

from . import exceptions
from . import utils

MESSAGE_MAX_LENGTH = 8


def signal_decode(signal_name, signal_info,
        message_binary_msb, message_binary_lsb, message_binary_length):
    """
    Decode a signal of a CAN message

    Arguments:
        signal_name: str, human name of the signal
        signal_info: dict, information about the signal (length, endianness)
        message_binary_msb: str, data of CAN message in MSB binary format
        message_binary_lsb: str, data of CAN message in LSB binary format
        message_binary_length: int, length of CAN message in binary format

    Returns:
        signal: dict, a decoded signal from the CAN data message

    Raises:
        exceptions.InvalidBitStart: when signal's bit start is too high
    """
    signal = {
        'name': signal_name,
        'length': signal_info['length']
    }
    bit_start = signal_info['bit_start']
    is_little_endian = bool(signal_info['little_endian'])
    signal['is_little_endian'] = is_little_endian

    if is_little_endian:
        message_binary = message_binary_lsb
    else:
        message_binary = message_binary_msb

    if bit_start >= message_binary_length:
        raise exceptions.InvalidBitStart("Bit start %d of signal %s is too high" % (
            bit_start, signal_name))

    if is_little_endian:
        # In Intel format (little-endian), bit_start is the position of the
        # Least Significant Bit so it needs to be byte swapped
        signal['bit_end'] = message_binary_length - bit_start
        signal['bit_start'] = signal['bit_end'] - signal['length']
    else:
        # Motorola. Weird thing of the DBC format
        signal['bit_start'] = (bit_start // 8) * 8 + (7 - (bit_start % 8))
        signal['bit_end'] = signal['bit_start'] + signal['length']

    s_value = message_binary[signal['bit_start']:signal['bit_end']]

    signal['factor'] = int(signal_info.get('factor', 1))
    signal['offset'] = int(signal_info.get('offset', 0))
    signal['value'] = int(s_value, 2) * signal['factor'] + signal['offset']
    signal['unit'] = signal_info.get('unit', '')

    return signal


def message_get_current_multiplexing_mode(message_info, message_binary_msb, message_binary_lsb,
        message_binary_length):
    """Extract the current multiplexing mode of the message if any.

    Arguments:
        message_info: dict, information about the message (name, signals)
        message_binary_msb: str, data of CAN message in MSB binary format
        message_binary_lsb: str, data of CAN message in LSB binary format
        message_binary_length: int, length of CAN message in binary format

    Returns:
        multiplexing_mode: int, the current multiplexing mode or None
    """
    multiplexing_mode = None
    signals = message_info['signals']
    for signal_name, signal_info in signals.items():
        if signal_info.get('multiplexor', False):
            signal = signal_decode(signal_name, signal_info, message_binary_msb, message_binary_lsb,
                message_binary_length)
            multiplexing_mode = signal['value']
            break

    return multiplexing_mode


def message_decode(message_id, message_length, message_data, dbc_json):
    """Decode a CAN message (also called a frame).

    Args:
        message_id: int, message identifier.
        message_length: int, length of the useful data in message data received (the length
            can be different of the CAN data length).
        message_data: bytes, binary data of the message.
        dbc_json: dict, deserialized version of a DBC file converted to JSON with libcanardbc.

    Returns:
        message: decoded message with list of signals.

    Raises:
        exceptions.InvalidDBC: when used DBC has not messages entry
        exceptions.MessageNotFound: when message's ID is not found in the DBC
    """
    if 'messages' not in dbc_json:
        raise exceptions.InvalidDBC("Invalid DBC file (no messages entry)")

    # Initialize the returns
    message = {'signals': []}

    try:
        message_info = dbc_json['messages'][str(message_id)]
        message['name'] = message_info['name']
        message['id'] = message_id
    except KeyError:
        raise exceptions.MessageNotFound("Message ID {id:d} (0x{id:x}) not found in DBC".format(id=message_id))

    # The CAN message data is always 8 bytes so it's required to truncate it to keep only
    # the useful bytes (nop when already truncated)
    message_data = message_data[:message_length]

    # Convert length from bytes to bits
    message_binary_length = message_length * 8

    # Motorola
    # 0n to fit in n characters width with 0 padding
    message_binary_msb = format(
        int.from_bytes(message_data, byteorder='big', signed=False),
        '0%db' % message_binary_length)

    # For Intel, identical but swapped
    message_binary_lsb = utils.swap_bytes(message_binary_msb)

    if message_info.get('has_multiplexor', False):
        multiplexing_mode = message_get_current_multiplexing_mode(
            message_info, message_binary_msb, message_binary_lsb,
            message_binary_length)
    else:
        multiplexing_mode = None
    message['multiplexing_mode'] = multiplexing_mode

    message['raw_data'] = message_data
    signals = sorted(message_info['signals'].items(), key=lambda t: int(t[1]['bit_start']))
    for signal_name, signal_info in signals:
        # Decode signal only when no multiplexor or the signal is associated to the
        # current mode or the signal is not multiplexed
        if (multiplexing_mode is None or
                multiplexing_mode == signal_info.get('multiplexing', multiplexing_mode)):
            signal = signal_decode(
                signal_name, signal_info, message_binary_msb, message_binary_lsb,
                message_binary_length)
            message['signals'].append(signal)

    return message



# Copyright © 2015 Polyconseil SAS
# SPDX-License-Identifier: BSD-3-Clause
#


class CanetonError(Exception):
    """Caneton base exception."""


class DecodingError(CanetonError):
    """Raised when there is a CAN frame decoding error."""


class InvalidBitStart(DecodingError):
    """Raised during decoding of a CAN signal, when the bit start value is invalid (e.g. too high)."""


class InvalidDBC(DecodingError):
    """Raised during decoding of a CAN message, when used DBC have missing sections."""


class MessageNotFound(DecodingError):
    """Raised during decoding of a CAN message, when identifier of this last is not found in the DBC."""



# Copyright © 2015 Polyconseil SAS
# SPDX-License-Identifier: BSD-3-Clause
#

from unittest import TestCase

import caneton
from caneton import cli


class TestCLI(TestCase):

    def setUp(self):
        self.parser = cli.create_parser()  # flake8: noqa

    def test_message_701(self):
        args = self.parser.parse_args(['./tests/dbc.json', '0x701', '0x01780178010000'])
        cleaned_args = cli.args_cleanup(args)
        args.dbcfile.close()
        message = caneton.message_decode(**cleaned_args)
        self.assertEqual(message['name'], 'CU_MULTI_FOO_BAR')
        signals = message['signals']
        expected_signals = [
            {'name': 'Mode', 'value': 1},
            {'name': 'Bar1', 'value': 376},
            {'name': 'Bar2', 'value': 376},
        ]
        self.assertEqual(len(signals), len(expected_signals))
        for (signal, expected_signal) in zip(signals, expected_signals):
            self.assertEqual(signal['name'], expected_signal['name'])
            self.assertEqual(signal['value'], expected_signal['value'], signal['name'])



#!/usr/bin/env python3
#
# Copyright © 2015 Polyconseil SAS
# SPDX-License-Identifier: BSD-3-Clause
#
# Interpret CAN message DLC (payload) to extract the values of the signals.
# Requires a DBC file.
#

import json
import argparse

import caneton


def create_parser():
    parser = argparse.ArgumentParser(description="Decode of CAN message with the help of DBC information.")
    parser.add_argument(
        'dbcfile', type=argparse.FileType('r'),
        help="DBC file converted in JSON format to use for decoding.")
    parser.add_argument('id', type=str, help="ID of the message on the CAN bus (eg. 0x5BB or 1467)")
    parser.add_argument('data', type=str, help="message data in hexadecimal (eg. 0x1112131415161718)")
    parser.add_argument('--output', type=str, choices=['json', 'text'], default='text',
        help="Format of the output (JSON or text)")
    return parser

def args_cleanup(args):
    # Check and cleanup message ID (minium 0x1)
    if len(args.id) > 2 and args.id[:2] == '0x':
        # It's an hexadecimal number
        message_id = int(args.id, 16)
    else:
        message_id = int(args.id)

    # Check the message data
    # Check the length of the data before removing the 0x prefix
    if len(args.data) < 4:
        raise ValueError("The CAN message data is too short '%s'." % args.data)

    # Test the first bytes are the 0x prefix
    if args.data[:2] != '0x':
        raise ValueError("The CAN message data '%s' is not prefixed by 0x." % args.data)

    # Remove the 0x prefix
    data = args.data[2:]
    data_length = len(data)

    is_multiple_of_two = (data_length % 2) == 0
    if not is_multiple_of_two:
        raise ValueError("The CAN data is not a multiple of two '%s'." % args.data)

    byte_length = data_length // 2
    if byte_length > caneton.MESSAGE_MAX_LENGTH:
        raise ValueError("The CAN message data length is too large (%d > %d)" % (
            byte_length, caneton.MESSAGE_MAX_LENGTH))

    try:
        # Convert hexadecimal string to bytes
        length, data = caneton.hex_ascii_to_bytes(data)
    except ValueError:
        raise ValueError("Invalid data argument '%s'." % args.data)

    # Load file as JSON file
    try:
        dbc_json = json.loads(args.dbcfile.read())
    except ValueError:
        raise ValueError("Unable to load the DBC file '%s' as JSON." % args.dbcfile)

    return {
        'message_id': message_id, 'message_data': data, 'message_length': length,
        'dbc_json': dbc_json
    }


def message_output(message, is_json_output):
    if is_json_output:
        return message
    else:
        print("Message {name} {id:d} (0x{id:x})".format(**message))
        if message['multiplexing_mode']:
            print("Multiplexing mode: %s" % message['multiplexing_mode'])

        for signal in message['signals']:
            signal['endianness'] = 'LSB' if signal['is_little_endian'] else 'MSB'
            print("Signal {name} - ({length}@{bit_start} {endianness})x{factor}+{offset} = {value} {unit}".format(
                **signal))

def main():
    parser = create_parser()
    args = parser.parse_args()
    cleaned_args = args_cleanup(args)
    args.dbcfile.close()

    message = caneton.message_decode(**cleaned_args)
    message_output(message, args.output == 'json')



if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# Copyright (c) Polyconseil SAS. All rights reserved.

from setuptools import find_packages, setup

VERSION = '1.0.2'

setup(
    name="caneton",
    version=VERSION,
    author="Polyconseil dev team",
    author_email="autolib-dev@polyconseil.fr",
    description=("Decode CAN messages"),
    license="BSD-3-Clause",
    keywords="CAN DBC",
    url="https://github.com/polyconseil/caneton",
    packages=find_packages(),
    long_description="",
    entry_points={
        'console_scripts': [
            'caneton-decode = caneton.cli:main',
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
    ],
    test_suite='',
    include_package_data=True,
)
